from myst.client import Client

from ...models.model_fit_result_list import ModelFitResultList


def request_sync(client: Client, model_uuid: str) -> ModelFitResultList:
    """Lists model fit results for the given model."""

    return client.request(
        method="get", path=f"/models/{model_uuid}/model_fit_results/", response_class=ModelFitResultList
    )



from myst.client import Client

from ...models.layer_list import LayerList


def request_sync(client: Client, time_series_uuid: str) -> LayerList:
    """Lists layers for a time series."""

    return client.request(method="get", path=f"/time_series/{time_series_uuid}/layers/", response_class=LayerList)



from typing import List

from myst.models import base_model

from ..models.model_connector_get import ModelConnectorGet


class ModelConnectorList(base_model.BaseModel):
    """Schema for model connector list responses."""

    data: List[ModelConnectorGet]
    has_more: bool



from myst.client import Client

from ...models.source_get import SourceGet
from ...models.source_update import SourceUpdate


def request_sync(client: Client, uuid: str, json_body: SourceUpdate) -> SourceGet:
    """Updates a source."""

    return client.request(method="patch", path=f"/sources/{uuid}", response_class=SourceGet, request_model=json_body)



from typing import Any, Dict, Tuple, Type, Union

import pydantic
from pydantic import BaseConfig, Field, root_validator
from pydantic.fields import FieldInfo, ModelField
from typing_extensions import Literal


class DiscriminatedFieldInfo(FieldInfo):
    @staticmethod
    def _get_field_info(
        field_name: str, annotation: Any, value: Any, config: Type["BaseConfig"]
    ) -> Tuple["DiscriminatedFieldInfo", Any]:
        return cls._get_field_info(field_name=field_name, annotation=annotation, value=value, config=config)

    # def __init__(self, field_info: FieldInfo, discriminator: str) -> None:
    #     self._field_info = field_info
    #     self._discriminator = discriminator


def discriminated_field(*args: Any, discriminator: str, **kwargs: Any) -> Any:
    field_info = Field(*args, **kwargs)
    return DiscriminatedFieldInfo(field_info=field_info, discriminator=discriminator)


# class DiscriminatedUnion:
#     @classmethod
#     def __get_validators(cls) -> Generator[Callable]:
#         yield cls.validate

#     @classmethod
#     def validate(cls, v: Any) -> Any:


class TimeSeries(pydantic.BaseModel):
    node_type: Literal["TimeSeries"]
    meows: int


class Operation(pydantic.BaseModel):
    node_type: Literal["Operation"]
    arfs: float


class NodeContainer(pydantic.BaseModel):
    node: Union[TimeSeries, Operation]  # = discriminated_field(..., discriminator="node_type")

    class Config:
        extra = pydantic.Extra.forbid

        discriminator = "node_type"
        discriminated_field = "node"

        @classmethod
        def prepare_field(cls, field: "ModelField") -> None:
            print(field)

        @classmethod
        def get_field_info(cls, name: str) -> Dict[str, Any]:
            return {}

    @root_validator(pre=True)
    def check_discriminator_present(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        discrim_field = cls.Config.discriminated_field
        assert discrim_field in values, f"missing discriminated field {discrim_field}"
        discrim_value = values[discrim_field]
        assert cls.Config.discriminator in discrim_value
        return values


NodeContainer(node={"node_type": "TimeSeries", "meows": 4})
NodeContainer(node={"node_type": "Operation", "meows": 4})
NodeContainer(node={"node_type": "TimeSeries", "arfs": 3.141596525837})
NodeContainer(node={"meows": 4})
NodeContainer(node={"random": "information"})



from myst.client import Client

from ...models.time_series_get import TimeSeriesGet
from ...models.time_series_update import TimeSeriesUpdate


def request_sync(client: Client, uuid: str, json_body: TimeSeriesUpdate) -> TimeSeriesGet:
    """Updates a time series."""

    return client.request(
        method="patch", path=f"/time_series/{uuid}", response_class=TimeSeriesGet, request_model=json_body
    )



from myst.client import Client

from ...models.model_fit_policy_list import ModelFitPolicyList


def request_sync(client: Client, model_uuid: str) -> ModelFitPolicyList:
    """Lists fit policies for a model."""

    return client.request(
        method="get", path=f"/models/{model_uuid}/model_fit_policies/", response_class=ModelFitPolicyList
    )



from typing import TYPE_CHECKING, Optional

from myst.adapters.timing import to_timing_create
from myst.client import get_client
from myst.models.timing import AbsoluteOrRelativeTiming, TimeRangeBoundary, from_time_range_boundary
from myst.models.types import ItemOrSlice
from myst.openapi.api.time_series import create_time_series_layer
from myst.openapi.models.layer_create import LayerCreate
from myst.resources.edge import Edge
from myst.resources.node import Node

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.time_series import TimeSeries


class Layer(Edge):
    """An edge into a time series.

    Layers are a way of stitching together data from multiple upstream nodes into a single, cohesive time series. Data
    can be combined across different time ranges, and a time series can use data from a lower-precedence layer when
    data is missing from a higher-precedence layer.

    Attributes:
        order: integer specifying priority of this layer when combining multiple layers; lower order implies higher
            precedence
        start_timing: the beginning of the natural time range this layer should produce data for; if None, there is no
            restriction on the beginning of the range
        end_timing: the end of the natural time range this layer should produce data for; if None, there is no
            restriction on the end of the range
    """

    order: int
    start_timing: Optional[AbsoluteOrRelativeTiming] = None
    end_timing: Optional[AbsoluteOrRelativeTiming] = None

    @classmethod
    def create(
        cls,
        downstream_node: "TimeSeries",
        upstream_node: Node,
        order: int,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
        start_timing: Optional[TimeRangeBoundary] = None,
        end_timing: Optional[TimeRangeBoundary] = None,
    ) -> "Layer":
        layer = create_time_series_layer.request_sync(
            client=get_client(),
            time_series_uuid=str(downstream_node.uuid),
            json_body=LayerCreate(
                object="Edge",
                type="Layer",
                upstream_node=str(upstream_node.uuid),
                order=order,
                output_index=output_index,
                label_indexer=label_indexer,
                start_timing=to_timing_create(from_time_range_boundary(start_timing)),
                end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            ),
        )

        return Layer.parse_obj(layer.dict())



from typing import List

from myst.models import base_model

from ..models.model_run_result_get import ModelRunResultGet


class ModelRunResultList(base_model.BaseModel):
    """Schema for model run result list responses."""

    data: List[ModelRunResultGet]
    has_more: bool



from myst.client import Client

from ...models.model_get import ModelGet
from ...models.model_update import ModelUpdate


def request_sync(client: Client, uuid: str, json_body: ModelUpdate) -> ModelGet:
    """Updates a model."""

    return client.request(method="patch", path=f"/models/{uuid}", response_class=ModelGet, request_model=json_body)



from typing import List

from myst.models import base_model

from ..models.operation_connector_get import OperationConnectorGet


class OperationConnectorList(base_model.BaseModel):
    """Schema for operation connector list responses."""

    data: List[OperationConnectorGet]
    has_more: bool



from typing import Any, Dict, Iterable, List, Optional, Sequence

import numpy as np
from pydantic import Field, root_validator, validator
from pydantic.typing import Literal

from myst.core.data.time_array import TimeArray
from myst.core.data.validators import (
    check_arrays_same_length_as_time_range,
    ensure_not_none,
    maybe_to_boolean_int_array,
    to_numpy_float_array,
)
from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta
from myst.models.base_model import BaseModel
from myst.models.types import AxisLabels, CoordinateLabels, Metadata, Shape


def _validate_time_array_compatibility(time_arrays: Iterable[TimeArray], check_shape: Optional[bool] = True) -> None:
    """Validates that the passed time arrays are compatible.

    Args:
        time_arrays: time arrays to validate
        check_shape: flag to indicate if we should check the time array shapes are equal

    Raises:
        ValueError: if the passed time arrays are incompatible
    """
    # Validate that the sample periods are the same across all time arrays.
    sample_period = None
    for time_array in time_arrays:
        if sample_period is not None:
            if time_array.sample_period != sample_period:
                raise ValueError("Time arrays do not share the same sample period.")
        else:
            sample_period = time_array.sample_period

    # Validate that the cell shapes are the same across all time arrays.
    cell_shape = None
    for time_array in time_arrays:
        if cell_shape is not None:
            if time_array.cell_shape != cell_shape:
                raise ValueError("Time arrays do not share the same cell shape.")
        else:
            cell_shape = time_array.cell_shape

    # Validate that the coordinate labels are the same across all time arrays.
    coordinate_labels = None
    for time_array in time_arrays:
        if coordinate_labels is not None:
            if not np.array_equal(time_array.coordinate_labels, coordinate_labels):
                raise ValueError("Time arrays do not share the same coordinate labels.")
        else:
            coordinate_labels = time_array.coordinate_labels

    # Validate that the axis labels are the same across all time arrays.
    axis_labels = None
    for time_array in time_arrays:
        if axis_labels is not None:
            if not np.array_equal(time_array.axis_labels, axis_labels):
                raise ValueError("Time arrays do not share the same axis labels.")
        else:
            axis_labels = time_array.axis_labels

    if check_shape:
        # Validate that the shapes are the same across all time arrays.
        shape = None
        for time_array in time_arrays:
            if shape is not None:
                if time_array.shape != shape:
                    raise ValueError("Time arrays do not share the same shape.")
            else:
                shape = time_array.shape


class TimeDatasetRow(BaseModel):
    start_time: Time = Field(
        ..., description="The time at which this time series data starts, inclusive.", example="2021-01-01T01:23:45Z"
    )
    end_time: Time = Field(
        ..., description="The time at which this time series data ends, exclusive.", example="2021-02-01T12:34:56Z"
    )
    as_of_time: Time = Field(
        ..., description="The time as of which this time series data was known.", example="2021-06-01T00:00:00Z"
    )
    values: np.ndarray = Field(
        ...,
        description="The multidimensional array of values for this time series.",
        examples=dict(
            empty=dict(summary="Empty array", value=[]),
            vector=dict(summary="1-D array of floats", value=[1.0, -47.47]),
            matrix=dict(summary="2-D array of floats", value=[[1.1, 1.2], [2.1, 2.2]]),
        ),
    )
    mask: Optional[np.ndarray] = Field(
        None,
        description=(
            "The mask to be used to filter out invalid or missing data. Must have the same shape as `values`, and each "
            "mask value must be either 0 or 1."
        ),
        examples=dict(
            empty=dict(summary="Empty array", value=[]),
            vector=dict(summary="1-D array of ints", value=[1, 0]),
            matrix=dict(summary="2-D array of ints", value=[[1, 0], [0, 1]]),
        ),
    )

    def dict(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """Returns the dictionary representation of this `TimeDatasetRow`."""
        dict_repr = super().dict(*args, **kwargs)

        arr: np.ndarray = dict_repr["values"]
        dict_repr["values"] = np.where(np.isnan(arr), None, arr).tolist()

        if self.mask is not None:
            dict_repr["mask"] = dict_repr["mask"].tolist()

        return dict_repr

    _ensure_values_not_none = validator("values", pre=True, allow_reuse=True)(ensure_not_none)
    _coerce_values_to_float_array = validator("values", pre=True, allow_reuse=True)(to_numpy_float_array)

    _coerce_mask_to_boolean_int_array = validator("mask", pre=True, allow_reuse=True)(maybe_to_boolean_int_array)

    class Config:
        # Allow arbitrary types in order to support `np.ndarray` directly.
        arbitrary_types_allowed = True


class TimeDataset(BaseModel):
    object: Literal["TimeDataset"] = Field(
        example="TimeDataset",
        description="The field representing the type of object this resource is. Must be `TimeDataset`.",
    )
    coordinate_labels: Optional[CoordinateLabels] = Field(
        example=("field",), description="The labels for each coordinate of a cell in the time dataset."
    )
    axis_labels: Optional[AxisLabels] = Field(
        example=(), description="The labels for each axis of a cell in the time dataset."
    )
    sample_period: TimeDelta = Field(
        example="PT1H", description="The sample period of the time dataset formatted as an ISO 8601 Duration string."
    )
    metadata: Optional[Metadata] = Field(
        example={"key": "value"}, description="User-defined metadata about the time dataset."
    )
    data: List[TimeDatasetRow] = Field(description="List of rows belonging to this time dataset.", min_items=1)
    cell_shape: Shape = Field(None, example=(), description="The shape of each cell in the time dataset.")

    @validator("data")
    def ensure_unique_as_of_times(cls, v: List[TimeDatasetRow]) -> List[TimeDatasetRow]:
        as_of_times = [row.as_of_time for row in v]
        if len(set(as_of_times)) != len(as_of_times):
            raise ValueError("The `as_of_time` for each `TimeDatasetRow` must be unique.")

        return v

    @validator("data")
    def check_cell_shapes_equal(cls, v: List[TimeDatasetRow]) -> List[TimeDatasetRow]:
        if len(set([row.values.shape[1:] for row in v])) > 1:
            raise ValueError("The cell shape of `values` in each `TimeDatasetRow` must match.")

        return v

    @validator("data")
    def sort_rows_by_as_of_time_desc(cls, v: List[TimeDatasetRow]) -> List[TimeDatasetRow]:
        return sorted(v, key=lambda row: row.as_of_time, reverse=True)

    @validator("cell_shape", always=True)
    def infer_cell_shape_from_data(cls, v: Any, values: Dict[str, Any]) -> Any:
        data = values.get("data", None)

        # We rely on previous validation to ensure that each row has the same cell shape.
        if v is None and data is not None and len(data):
            return data[0].values.shape[1:]

        return v

    @root_validator
    def validate_time_range_consistent_with_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validates that the given time range is consistent with the values and mask."""
        sample_period = values.get("sample_period", None)

        if sample_period is not None:
            for row in values.get("data", ()):
                if row.start_time and row.end_time and row.values is not None:
                    check_arrays_same_length_as_time_range(
                        start_time=row.start_time,
                        end_time=row.end_time,
                        sample_period=sample_period,
                        values=row.values,
                        mask=row.mask,
                    )

        return values

    @classmethod
    def from_time_arrays(cls, time_arrays: Sequence[TimeArray]) -> "TimeDataset":
        if not time_arrays:
            raise ValueError("Cannot construct `TimeDataset` from empty collection.")

        _validate_time_array_compatibility(time_arrays=time_arrays, check_shape=False)

        first_time_array = time_arrays[0]

        time_dataset_rows = [
            TimeDatasetRow(
                start_time=time_array.start_time,
                end_time=time_array.end_time,
                as_of_time=time_array.as_of_time,
                values=time_array.values,
                mask=time_array.mask,
            )
            for time_array in time_arrays
        ]

        return TimeDataset(
            object="TimeDataset",
            cell_shape=first_time_array.cell_shape,
            sample_period=first_time_array.sample_period,
            data=time_dataset_rows,
            coordinate_labels=first_time_array.coordinate_labels,
            axis_labels=first_time_array.axis_labels,
        )

    def to_time_arrays(self) -> List[TimeArray]:
        """Returns the data in this time dataset as a list of time arrays."""
        return [
            TimeArray(
                sample_period=self.sample_period,
                start_time=time_dataset_row.start_time,
                end_time=time_dataset_row.end_time,
                as_of_time=time_dataset_row.as_of_time,
                values=time_dataset_row.values,
                mask=time_dataset_row.mask,
                coordinate_labels=self.coordinate_labels,
                axis_labels=self.axis_labels,
            )
            for time_dataset_row in self.data
        ]



from myst.client import Client

from ...models.input_create import InputCreate
from ...models.input_get import InputGet


def request_sync(client: Client, operation_uuid: str, json_body: InputCreate) -> InputGet:
    """Creates an input for an operation."""

    return client.request(
        method="post", path=f"/operations/{operation_uuid}/inputs/", response_class=InputGet, request_model=json_body
    )



from myst.client import Client

from ...models.model_fit_policy_get import ModelFitPolicyGet
from ...models.model_fit_policy_update import ModelFitPolicyUpdate


def request_sync(client: Client, model_uuid: str, uuid: str, json_body: ModelFitPolicyUpdate) -> ModelFitPolicyGet:
    """Updates a model fit policy."""

    return client.request(
        method="patch",
        path=f"/models/{model_uuid}/model_fit_policies/{uuid}",
        response_class=ModelFitPolicyGet,
        request_model=json_body,
    )



from myst.client import Client

from ...models.source_run_result_list import SourceRunResultList


def request_sync(client: Client, source_uuid: str) -> SourceRunResultList:
    """Lists source run results for the given source."""

    return client.request(
        method="get", path=f"/sources/{source_uuid}/source_run_results/", response_class=SourceRunResultList
    )



from uuid import UUID

from myst.models.timing import AbsoluteOrRelativeTiming
from myst.resources.resource import Resource


class Policy(Resource):
    """Describes when and over what natural time range to run a particular type of job for a node.

    Attributes:
        creator: the identifier of the user who created this resource
        schedule_timing: when the policy is scheduled to run, whether recurrent or once
        active: whether this policy is currently in effect
        node: creator: the identifier of the node this policy applies to
        start_timing: the beginning of the natural time range for which this policy applies, inclusive
        end_timing: the end of the natural time range for which this policy applies, exclusive
    """

    creator: UUID
    schedule_timing: AbsoluteOrRelativeTiming
    active: bool
    node: UUID
    start_timing: AbsoluteOrRelativeTiming
    end_timing: AbsoluteOrRelativeTiming


class ModelFitPolicy(Policy):
    """Describes when and over what natural time range to run a fit job on a model."""


class TimeSeriesRunPolicy(Policy):
    """Describes when and over what natural time range to run a fit job on a model."""



from myst.client import Client

from ...models.operation_connector_parameters_schema_get import OperationConnectorParametersSchemaGet


def request_sync(client: Client, uuid: str) -> OperationConnectorParametersSchemaGet:
    """Gets an operation connector's parameters schema."""

    return client.request(
        method="get",
        path=f"/operation_connectors/{uuid}:get_parameters_schema",
        response_class=OperationConnectorParametersSchemaGet,
    )



from myst.client import Client

from ...models.layer_create import LayerCreate
from ...models.layer_get import LayerGet


def request_sync(client: Client, time_series_uuid: str, json_body: LayerCreate) -> LayerGet:
    """Creates a layer for a time series."""

    return client.request(
        method="post", path=f"/time_series/{time_series_uuid}/layers/", response_class=LayerGet, request_model=json_body
    )



from myst.client import Client

from ...models.source_list import SourceList


def request_sync(client: Client) -> SourceList:
    """Lists sources."""

    return client.request(method="get", path=f"/sources/", response_class=SourceList)



from myst.client import Client

from ...models.model_fit_policy_create import ModelFitPolicyCreate
from ...models.model_fit_policy_get import ModelFitPolicyGet


def request_sync(client: Client, model_uuid: str, json_body: ModelFitPolicyCreate) -> ModelFitPolicyGet:
    """Creates a model fit policy."""

    return client.request(
        method="post",
        path=f"/models/{model_uuid}/model_fit_policies/",
        response_class=ModelFitPolicyGet,
        request_model=json_body,
    )



from typing import List

from myst.models import base_model

from ..models.time_dataset_spec import TimeDatasetSpec


class OutputSpecsGet(base_model.BaseModel):
    """Schema for node output spec responses."""

    output_specs: List[TimeDatasetSpec]



from typing import TYPE_CHECKING, Optional, Union

from myst.client import Client, get_client
from myst.models.types import ItemOrSlice
from myst.openapi.api.models import create_model_input
from myst.openapi.api.operations import create_operation_input
from myst.openapi.models.input_create import InputCreate
from myst.openapi.models.input_get import InputGet
from myst.resources.edge import Edge
from myst.resources.time_series import TimeSeries

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.model import Model
    from myst.resources.operation import Operation


def _create_input(downstream_node: Union["Model", "Operation"], input_create: InputCreate, client: Client) -> InputGet:
    from myst.resources.model import Model  # Avoid circular imports.

    if isinstance(downstream_node, Model):
        return create_model_input.request_sync(
            client=client, model_uuid=str(downstream_node.uuid), json_body=input_create
        )
    else:  # i.e. `to` is an `Operation`
        return create_operation_input.request_sync(
            client=client, operation_uuid=str(downstream_node.uuid), json_body=input_create
        )


class Input(Edge):
    """An edge from a time series into a model or operation."""

    @classmethod
    def create(
        cls,
        downstream_node: Union["Model", "Operation"],
        upstream_node: TimeSeries,
        group_name: str,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
    ) -> "Input":
        """Creates a new input edge between the given nodes.

        Args:
            downstream_node: the model or operation into which data flows out of this edge
            upstream_node: the time series from which data flows into this edge
            group_name: the name of the group of inputs on the underlying model or operation connector to which to pass
                the data from this input
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to the downstream
                node
            label_indexer: the slice of the upstream data to pass to the downstream node

        Returns:
            the newly created input
        """
        input_create = InputCreate(
            object="Edge",
            type="Input",
            upstream_node=str(upstream_node.uuid),
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

        input_ = _create_input(downstream_node=downstream_node, input_create=input_create, client=get_client())

        return Input.parse_obj(input_.dict())



from myst.client import Client

from ...models.model_connector_get import ModelConnectorGet


def request_sync(client: Client, uuid: str) -> ModelConnectorGet:
    """Gets a model connector by its unique identifier."""

    return client.request(method="get", path=f"/model_connectors/{uuid}", response_class=ModelConnectorGet)



from myst.client import Client

from ...models.operation_connector_list import OperationConnectorList


def request_sync(client: Client) -> OperationConnectorList:
    """Lists operation connectors."""

    return client.request(method="get", path=f"/operation_connectors/", response_class=OperationConnectorList)



from typing import Any, Dict, Optional

from myst.client import Client

from ...models.time_series_query_result_get import TimeSeriesQueryResultGet


def request_sync(
    client: Client, uuid: str, start_time: str, end_time: str, as_of_time: Optional[str] = None
) -> TimeSeriesQueryResultGet:
    """Queries time series data."""

    params: Dict[str, Any] = {"start_time": start_time, "end_time": end_time, "as_of_time": as_of_time}
    params = {k: v for k, v in params.items() if v is not None}

    return client.request(
        method="get", path=f"/time_series/{uuid}:query", response_class=TimeSeriesQueryResultGet, params=params
    )



from myst.client import Client

from ...models.time_series_run_result_get import TimeSeriesRunResultGet


def request_sync(client: Client, time_series_uuid: str, uuid: str) -> TimeSeriesRunResultGet:
    """Gets a time series run result by its unique identifier."""

    return client.request(
        method="get",
        path=f"/time_series/{time_series_uuid}/time_series_run_results/{uuid}",
        response_class=TimeSeriesRunResultGet,
    )



from myst.client import Client

from ...models.input_specs_schema_get import InputSpecsSchemaGet


def request_sync(client: Client, uuid: str) -> InputSpecsSchemaGet:
    """Returns the JSON input specs schema for the specified node."""

    return client.request(
        method="get", path=f"/models/{uuid}:get_input_specs_schema", response_class=InputSpecsSchemaGet
    )



from typing import List

from myst.models import base_model

from ..models.operation_get import OperationGet


class OperationList(base_model.BaseModel):
    """Schema for operation list responses."""

    data: List[OperationGet]
    has_more: bool



from myst.client import Client

from ...models.time_series_insert import TimeSeriesInsert
from ...models.time_series_insert_result_get import TimeSeriesInsertResultGet


def request_sync(client: Client, uuid: str, json_body: TimeSeriesInsert) -> TimeSeriesInsertResultGet:
    """Inserts time series data."""

    return client.request(
        method="post",
        path=f"/time_series/{uuid}:insert",
        response_class=TimeSeriesInsertResultGet,
        request_model=json_body,
    )



from myst.client import Client

from ...models.operation_list import OperationList


def request_sync(client: Client) -> OperationList:
    """Lists operations."""

    return client.request(method="get", path=f"/operations/", response_class=OperationList)



from myst.client import Client

from ...models.source_connector_parameters_schema_get import SourceConnectorParametersSchemaGet


def request_sync(client: Client, uuid: str) -> SourceConnectorParametersSchemaGet:
    """Gets a source connector's parameters schema."""

    return client.request(
        method="get",
        path=f"/source_connectors/{uuid}:get_parameters_schema",
        response_class=SourceConnectorParametersSchemaGet,
    )



from typing import Any, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class OperationCreate(base_model.BaseModel):
    """Schema for operation create requests."""

    object_: Literal["Node"] = Field(..., alias="object")
    title: str
    project: str
    connector_uuid: str
    type: Optional[Literal["Operation"]] = "Operation"
    description: Optional[str] = None
    parameters: Optional[Any] = None



from myst.client import Client

from ...models.source_connector_list import SourceConnectorList


def request_sync(client: Client) -> SourceConnectorList:
    """Lists source connectors."""

    return client.request(method="get", path=f"/source_connectors/", response_class=SourceConnectorList)



from myst.client import Client

from ...models.model_fit_policy_get import ModelFitPolicyGet


def request_sync(client: Client, model_uuid: str, uuid: str) -> ModelFitPolicyGet:
    """Deletes a model fit policy."""

    return client.request(
        method="delete", path=f"/models/{model_uuid}/model_fit_policies/{uuid}", response_class=ModelFitPolicyGet
    )



from myst.client import Client

from ...models.input_get import InputGet


def request_sync(client: Client, model_uuid: str, uuid: str) -> InputGet:
    """Deletes a model input."""

    return client.request(method="delete", path=f"/models/{model_uuid}/inputs/{uuid}", response_class=InputGet)



from typing import TYPE_CHECKING, List, Optional

from myst.client import get_client
from myst.connectors.operation_connector import OperationConnector
from myst.models.types import ItemOrSlice, UUIDOrStr, to_uuid
from myst.openapi.api.operations import create_operation, get_operation, list_operation_inputs
from myst.openapi.models.operation_create import OperationCreate
from myst.resources.connector_node import ConnectorNode
from myst.resources.input import Input
from myst.resources.time_series import TimeSeries

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.project import Project


class Operation(ConnectorNode):
    """A node that performs a specified transformation on its input.

    In contrast to a model, an operation has no training phase and can only be run.
    """

    @classmethod
    def create(
        cls, project: "Project", title: str, connector: OperationConnector, description: Optional[str] = None
    ) -> "Operation":
        """Creates a new operation node.

        Args:
            project: the project in which to create the operation
            title: the title of the operation
            connector: the operation connector to use in the operation node
            description: a brief description of the operation

        Returns:
            the newly created operation
        """
        operation = create_operation.request_sync(
            client=get_client(),
            json_body=OperationCreate(
                object="Node",
                type="Operation",
                title=title,
                description=description,
                project=str(project.uuid),
                connector_uuid=str(connector.uuid),
                parameters=connector.parameters_exclude_none(),
            ),
        )

        return Operation.parse_obj(operation.dict())

    @classmethod
    def get(cls, uuid: UUIDOrStr) -> "Operation":
        """Gets a specific operation by its identifier."""
        operation = get_operation.request_sync(client=get_client(), uuid=str(to_uuid(uuid)))

        return Operation.parse_obj(operation.dict())

    def create_input(
        self,
        upstream_node: TimeSeries,
        group_name: str,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
    ) -> Input:
        """Creates an input into this operation.

        Args:
            upstream_node: the time series to feed into this operation
            group_name: the name of the input group on this operation's connector to which to pass the data from this
                input
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to this operation
            label_indexer: the slice of the upstream data to pass to this operation

        Returns:
            the newly created input
        """
        return Input.create(
            downstream_node=self,
            upstream_node=upstream_node,
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

    def list_inputs(self) -> List[Input]:
        """Lists all inputs into this operation."""
        operation_inputs = list_operation_inputs.request_sync(client=get_client(), operation_uuid=str(self.uuid))

        return [Input.parse_obj(input_.dict()) for input_ in operation_inputs.data]



from typing import List, Union

from myst.models import base_model

from ..models.model_get import ModelGet
from ..models.operation_get import OperationGet
from ..models.source_get import SourceGet
from ..models.time_series_get import TimeSeriesGet


class ProjectNodesList(base_model.BaseModel):
    """Project nodes list schema."""

    data: List[Union[TimeSeriesGet, SourceGet, OperationGet, ModelGet]]
    has_more: bool



from myst.client import Client

from ...models.time_series_list import TimeSeriesList


def request_sync(client: Client) -> TimeSeriesList:
    """Lists time series."""

    return client.request(method="get", path=f"/time_series/", response_class=TimeSeriesList)



from typing import List

from myst.models import base_model

from ..models.model_fit_result_get import ModelFitResultGet


class ModelFitResultList(base_model.BaseModel):
    """Schema for model fit result list responses."""

    data: List[ModelFitResultGet]
    has_more: bool



from myst.client import Client

from ...models.operation_create import OperationCreate
from ...models.operation_get import OperationGet


def request_sync(client: Client, json_body: OperationCreate) -> OperationGet:
    """Creates an operation."""

    return client.request(method="post", path=f"/operations/", response_class=OperationGet, request_model=json_body)



import enum
from dataclasses import dataclass
from typing import List
from uuid import UUID

from myst.connectors.source_connector import SourceConnector


@enum.unique
class YesEnergyAggregation(str, enum.Enum):
    AVG = "AVG"
    MAX = "MAX"
    MIN = "MIN"


@dataclass
class YesEnergyItem:
    datatype: str
    object_id: int


class YesEnergy(SourceConnector):
    def __init__(
        self,
        username: str,
        password: str,
        items: List[YesEnergyItem],
        stat: YesEnergyAggregation = YesEnergyAggregation.AVG,
    ) -> None:
        super().__init__(
            uuid=UUID("207da43c-d284-44ee-90d0-61b96fa4df1c"),
            parameters=dict(username=username, password=password, items=items, stat=stat),
        )



from myst.client import Client

from ...models.source_run_result_get import SourceRunResultGet


def request_sync(client: Client, source_uuid: str, uuid: str) -> SourceRunResultGet:
    """Gets a source run result by its unique identifier."""

    return client.request(
        method="get", path=f"/sources/{source_uuid}/source_run_results/{uuid}", response_class=SourceRunResultGet
    )



from typing import List

from myst.models import base_model

from ..models.operation_run_result_get import OperationRunResultGet


class OperationRunResultList(base_model.BaseModel):
    """Schema for operation run result list responses."""

    data: List[OperationRunResultGet]
    has_more: bool



import enum
from typing import Mapping, NamedTuple, Optional

from myst.connectors.source_connectors import cleaned_observations, enhanced_forecast, historical_hourly_conditions
from myst.core.time.time_delta import TimeDelta
from myst.models.types import AxisLabels, CoordinateLabels, Shape
from myst.recipes.time_series_recipes.time_series_recipe import TimeSeriesRecipe
from myst.resources.layer import Layer
from myst.resources.project import Project
from myst.resources.source import Source
from myst.resources.time_series import TimeSeries

_HISTORICAL_APIS_CUTOVER_BOUNDARY = TimeDelta("-PT23H")
_HISTORICAL_TO_FORECAST_CUTOVER_BOUNDARY = TimeDelta("PT1H")


@enum.unique
class Field(enum.Enum):
    TEMPERATURE = enum.auto()
    RELATIVE_HUMIDITY = enum.auto()
    WIND_DIRECTION = enum.auto()
    WIND_SPEED = enum.auto()


_CLEANED_OBSERVATIONS_FIELD_MAPPINGS: Mapping[Field, cleaned_observations.Field] = {
    Field.TEMPERATURE: cleaned_observations.Field.SURFACE_TEMPERATURE_CELSIUS,
    Field.RELATIVE_HUMIDITY: cleaned_observations.Field.RELATIVE_HUMIDITY_PERCENT,
    Field.WIND_DIRECTION: cleaned_observations.Field.WIND_DIRECTION_DEGREES,
    Field.WIND_SPEED: cleaned_observations.Field.WIND_SPEED_MPH,
}
_HISTORICAL_HOURLY_CONDITIONS_FIELD_MAPPINGS: Mapping[Field, historical_hourly_conditions.Field] = {
    Field.TEMPERATURE: historical_hourly_conditions.Field.TEMPERATURE,
    Field.RELATIVE_HUMIDITY: historical_hourly_conditions.Field.RELATIVE_HUMIDITY,
    Field.WIND_DIRECTION: historical_hourly_conditions.Field.WIND_DIRECTION,
    Field.WIND_SPEED: historical_hourly_conditions.Field.WIND_SPEED,
}
_ENHANCED_FORECAST_FIELD_MAPPINGS = {
    Field.TEMPERATURE: enhanced_forecast.Field.TEMPERATURE,
    Field.RELATIVE_HUMIDITY: enhanced_forecast.Field.RELATIVE_HUMIDITY,
    Field.WIND_DIRECTION: enhanced_forecast.Field.WIND_DIRECTION,
    Field.WIND_SPEED: enhanced_forecast.Field.WIND_SPEED,
    # TODO(sjespersen): Do we need others?
}


class Location(NamedTuple):
    latitude: float
    longitude: float


@enum.unique
class MetarStation(Location, enum.Enum):
    # TODO(sjespersen): Get the right lat/lngs for these!
    KSFO = Location(latitude=37.618331, longitude=-122.378647)
    KSKX = Location(latitude=36.458937, longitude=-105.671378)


class TheWeatherCompany(TimeSeriesRecipe):
    def __init__(self, field: Field, metar_station: MetarStation) -> None:
        self._field = field
        self._metar_station = metar_station

    def create_time_series(
        self,
        project: Project,
        title: str,
        sample_period: TimeDelta,
        cell_shape: Shape = (),
        coordinate_labels: CoordinateLabels = (),
        axis_labels: AxisLabels = (),
        description: Optional[str] = None,
    ) -> TimeSeries:
        latitude = self._metar_station.latitude
        longitude = self._metar_station.longitude
        cleaned_observations_source = Source.create(
            project=project,
            title=f"Cleaned Observations ({self._metar_station.name})",
            connector=cleaned_observations.CleanedObservations(
                latitude=latitude, longitude=longitude, fields=[_CLEANED_OBSERVATIONS_FIELD_MAPPINGS[self._field]]
            ),
        )
        historical_hourly_conditions_source = Source.create(
            project=project,
            title=f"Historical Hourly Conditions ({self._metar_station.name})",
            connector=historical_hourly_conditions.HistoricalHourlyConditions(
                latitude=latitude,
                longitude=longitude,
                fields=[_HISTORICAL_HOURLY_CONDITIONS_FIELD_MAPPINGS[self._field]],
            ),
        )
        enhanced_forecast_source = Source.create(
            project=project,
            title=f"Enhanced Forecast ({self._metar_station.name})",
            connector=enhanced_forecast.EnhancedForecast(
                latitude=latitude, longitude=longitude, fields=[_ENHANCED_FORECAST_FIELD_MAPPINGS[self._field]]
            ),
        )
        time_series = TimeSeries.create(
            project=project,
            title=title,
            sample_period=sample_period,
            cell_shape=cell_shape,
            coordinate_labels=coordinate_labels,
            axis_labels=axis_labels,
            description=description,
        )
        Layer.create(
            downstream_node=time_series,
            upstream_node=cleaned_observations_source,
            order=0,
            output_index=0,
            label_indexer=_CLEANED_OBSERVATIONS_FIELD_MAPPINGS[self._field],
            end_timing=_HISTORICAL_APIS_CUTOVER_BOUNDARY,
        )
        Layer.create(
            downstream_node=time_series,
            upstream_node=historical_hourly_conditions_source,
            order=1,
            output_index=0,
            label_indexer=_HISTORICAL_HOURLY_CONDITIONS_FIELD_MAPPINGS[self._field],
            start_timing=_HISTORICAL_APIS_CUTOVER_BOUNDARY,
            end_timing=_HISTORICAL_TO_FORECAST_CUTOVER_BOUNDARY,
        )
        Layer.create(
            downstream_node=time_series,
            upstream_node=enhanced_forecast_source,
            order=2,
            output_index=0,
            label_indexer=_ENHANCED_FORECAST_FIELD_MAPPINGS[self._field],
            start_timing=_HISTORICAL_TO_FORECAST_CUTOVER_BOUNDARY,
        )
        return time_series



from myst.client import Client

from ...models.input_get import InputGet


def request_sync(client: Client, operation_uuid: str, uuid: str) -> InputGet:
    """Gets a specific input for an operation."""

    return client.request(method="get", path=f"/operations/{operation_uuid}/inputs/{uuid}", response_class=InputGet)



from typing import List

from myst.models import base_model

from ..models.source_connector_get import SourceConnectorGet


class SourceConnectorList(base_model.BaseModel):
    """Schema for source connector list responses."""

    data: List[SourceConnectorGet]
    has_more: bool



from myst.client import Client

from ...models.time_series_run_policy_get import TimeSeriesRunPolicyGet
from ...models.time_series_run_policy_update import TimeSeriesRunPolicyUpdate


def request_sync(
    client: Client, time_series_uuid: str, uuid: str, json_body: TimeSeriesRunPolicyUpdate
) -> TimeSeriesRunPolicyGet:
    """Updates a time series run policy."""

    return client.request(
        method="patch",
        path=f"/time_series/{time_series_uuid}/time_series_run_policies/{uuid}",
        response_class=TimeSeriesRunPolicyGet,
        request_model=json_body,
    )



from typing import Any, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model

from ..models.deploy_status import DeployStatus


class OperationGet(base_model.BaseModel):
    """Schema for operation get responses."""

    object_: Literal["Node"] = Field(..., alias="object")
    uuid: str
    create_time: str
    organization: str
    owner: str
    type: Literal["Operation"]
    title: str
    project: str
    creator: str
    deploy_status: DeployStatus
    connector_uuid: str
    parameters: Any
    update_time: Optional[str] = None
    description: Optional[str] = None



from myst.client import Client

from ...models.operation_run_result_get import OperationRunResultGet


def request_sync(client: Client, operation_uuid: str, uuid: str) -> OperationRunResultGet:
    """Gets an operation run result by its unique identifier."""

    return client.request(
        method="get",
        path=f"/operations/{operation_uuid}/operation_run_results/{uuid}",
        response_class=OperationRunResultGet,
    )



from myst.client import Client

from ...models.time_series_get import TimeSeriesGet


def request_sync(client: Client, uuid: str) -> TimeSeriesGet:
    """Deletes a new time series."""

    return client.request(method="delete", path=f"/time_series/{uuid}", response_class=TimeSeriesGet)



from myst.client import Client

from ...models.project_job_list import ProjectJobList


def request_sync(client: Client, project_uuid: str) -> ProjectJobList:
    """Lists jobs."""

    return client.request(method="get", path=f"/projects/{project_uuid}/jobs/", response_class=ProjectJobList)



from myst.client import Client

from ...models.source_get import SourceGet


def request_sync(client: Client, uuid: str) -> SourceGet:
    """Deletes a new source."""

    return client.request(method="delete", path=f"/sources/{uuid}", response_class=SourceGet)



from myst.client import Client

from ...models.output_specs_get import OutputSpecsGet


def request_sync(client: Client, project_uuid: str, node_uuid: str) -> OutputSpecsGet:
    """Returns the output specs for the specified node."""

    return client.request(
        method="get", path=f"/projects/{project_uuid}/nodes/{node_uuid}:get_output_specs", response_class=OutputSpecsGet
    )



from myst.client import Client

from ...models.model_connector_list import ModelConnectorList


def request_sync(client: Client) -> ModelConnectorList:
    """Lists model connectors."""

    return client.request(method="get", path=f"/model_connectors/", response_class=ModelConnectorList)



from typing import List

from myst.models import base_model

from ..models.source_get import SourceGet


class SourceList(base_model.BaseModel):
    """Schema for source list responses."""

    data: List[SourceGet]
    has_more: bool



from myst.client import Client

from ...models.model_connector_parameters_schema_get import ModelConnectorParametersSchemaGet


def request_sync(client: Client, uuid: str) -> ModelConnectorParametersSchemaGet:
    """Gets a model connector's parameters schema."""

    return client.request(
        method="get",
        path=f"/model_connectors/{uuid}:get_parameters_schema",
        response_class=ModelConnectorParametersSchemaGet,
    )



from myst.client import Client

from ...models.project_result_list import ProjectResultList


def request_sync(client: Client, project_uuid: str) -> ProjectResultList:
    """Lists results."""

    return client.request(method="get", path=f"/projects/{project_uuid}/results/", response_class=ProjectResultList)



from myst.client import Client

from ...models.layer_get import LayerGet


def request_sync(client: Client, time_series_uuid: str, uuid: str) -> LayerGet:
    """Deletes a layer."""

    return client.request(
        method="delete", path=f"/time_series/{time_series_uuid}/layers/{uuid}", response_class=LayerGet
    )



from myst.client import Client

from ...models.operation_run_result_list import OperationRunResultList


def request_sync(client: Client, operation_uuid: str) -> OperationRunResultList:
    """Lists operation run results for the given operation."""

    return client.request(
        method="get", path=f"/operations/{operation_uuid}/operation_run_results/", response_class=OperationRunResultList
    )



from typing import List

from myst.models import base_model

from ..models.source_run_result_get import SourceRunResultGet


class SourceRunResultList(base_model.BaseModel):
    """Schema for source run result list responses."""

    data: List[SourceRunResultGet]
    has_more: bool



from myst.client import Client

from ...models.source_connector_get import SourceConnectorGet


def request_sync(client: Client, uuid: str) -> SourceConnectorGet:
    """Gets a source connector by its unique identifier."""

    return client.request(method="get", path=f"/source_connectors/{uuid}", response_class=SourceConnectorGet)



from myst.client import Client

from ...models.input_get import InputGet
from ...models.input_update import InputUpdate


def request_sync(client: Client, model_uuid: str, uuid: str, json_body: InputUpdate) -> InputGet:
    """Updates an existing input for a model."""

    return client.request(
        method="patch", path=f"/models/{model_uuid}/inputs/{uuid}", response_class=InputGet, request_model=json_body
    )



from typing import TYPE_CHECKING, List, Optional, Union

from myst import get_client
from myst.adapters.timing import to_timing_create
from myst.core.data.time_array import TimeArray
from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta
from myst.models.time_dataset import TimeDataset, TimeDatasetRow
from myst.models.timing import ScheduleTiming, TimeRangeBoundary, from_schedule_specifier, from_time_range_boundary
from myst.models.types import AxisLabels, CoordinateLabels, ItemOrSlice, Shape, UUIDOrStr, to_uuid
from myst.openapi.api.time_series import (
    create_time_series,
    create_time_series_run_policy,
    get_time_series,
    insert_time_series_data,
    list_time_series_layers,
    list_time_series_run_policies,
    query_time_series_data,
)
from myst.openapi.models.time_series_create import TimeSeriesCreate
from myst.openapi.models.time_series_insert import TimeSeriesInsert
from myst.openapi.models.time_series_run_policy_create import TimeSeriesRunPolicyCreate
from myst.resources.layer import Layer
from myst.resources.node import Node
from myst.resources.policy import TimeSeriesRunPolicy

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.project import Project


class TimeSeries(Node):
    """A queryable, shareable node containing fixed-frequency, arbitrarily dimensioned time series data.

    A time series stitches together data from multiple layers into a single stream, by consuming data from lower-
    precedence layers when data from higher-precedence layers is not available, and by consuming only the upstream data
    that is within the time range bounds set on each layer.

    In addition to queries, a time series also supports inserts, which take precedence over data flowing through the
    input layers.

    Attributes:
        sample_period: the frequency at which the data is sampled
        cell_shape: the shape of the data in each cell
        coordinate_labels: the labels for each coordinate of a cell
        axis_labels: the labels for each axis of a cell
    """

    sample_period: TimeDelta
    cell_shape: Shape
    coordinate_labels: CoordinateLabels
    axis_labels: AxisLabels

    @classmethod
    def create(
        cls,
        project: "Project",
        title: str,
        sample_period: TimeDelta,
        cell_shape: Shape = (),
        coordinate_labels: CoordinateLabels = (),
        axis_labels: AxisLabels = (),
        description: Optional[str] = None,
    ) -> "TimeSeries":
        """Creates a new time series node.

        Args:
            project: the project in which to create the time series
            title: the title of the time series
            sample_period: the frequency at which the data is sampled
            cell_shape: the shape of the data in each cell of the time series
            coordinate_labels: the labels for each coordinate of a cell in the time series
            axis_labels: the labels for each axis of a cell in the time series
            description: a brief description of the time series

        Returns:
            the newly created time series
        """
        time_series = create_time_series.request_sync(
            client=get_client(),
            json_body=TimeSeriesCreate(
                object="Node",
                type="TimeSeries",
                title=title,
                project=str(project.uuid),
                sample_period=sample_period.to_iso_string(),
                cell_shape=cell_shape,
                coordinate_labels=coordinate_labels,
                axis_labels=axis_labels,
                description=description,
            ),
        )

        return TimeSeries.parse_obj(time_series.dict())

    @classmethod
    def get(cls, uuid: UUIDOrStr) -> "TimeSeries":
        """Gets a specific time series by its identifier."""
        time_series = get_time_series.request_sync(client=get_client(), uuid=str(to_uuid(uuid)))

        return TimeSeries.parse_obj(time_series.dict())

    def insert_time_array(self, time_array: TimeArray) -> None:
        """Inserts data from a `TimeArray` into this time series.

        The data from this manual insertion will take precedence over the other layers flowing into this time series.

        Args:
            time_array: the array of data to be inserted
        """
        time_dataset_row = TimeDatasetRow(
            start_time=time_array.start_time,
            end_time=time_array.end_time,
            as_of_time=time_array.as_of_time,
            values=time_array.values,
            mask=time_array.mask,
        )

        time_dataset = TimeDataset(
            object="TimeDataset",
            cell_shape=time_array.cell_shape,
            sample_period=time_array.sample_period,
            data=[time_dataset_row],
            coordinate_labels=time_array.coordinate_labels,
            axis_labels=time_array.axis_labels,
        )

        insert_time_series_data.request_sync(
            client=get_client(), uuid=str(self.uuid), json_body=TimeSeriesInsert(time_dataset=time_dataset)
        )

    def query_time_array(self, start_time: Time, end_time: Time, as_of_time: Optional[Time] = None) -> TimeArray:
        """Queries this time series for data according to the given parameters.

        Args:
            start_time: the beginning of the natural time range to query over, inclusive
            end_time: the end of the natural time range to query over, exclusive
            as_of_time: the precise as of time to query

        Returns:
            a time array containing data for the specified time range and as of time

        Raises:
            NotImplementedError: if the underlying `TimeDataset` contains multiple `TimeArray` objects
        """
        time_series_query_result_get = query_time_series_data.request_sync(
            client=get_client(),
            uuid=str(self.uuid),
            start_time=start_time.to_iso_string(),
            end_time=end_time.to_iso_string(),
            as_of_time=(as_of_time and as_of_time.to_iso_string()) or None,
        )

        time_arrays = TimeDataset.parse_obj(time_series_query_result_get.time_dataset).to_time_arrays()

        # Without offering the user the ability to control flattening behavior, we just throw up our hands instead.
        if len(time_arrays) != 1:
            raise NotImplementedError()

        return time_arrays[0]

    def create_layer(
        self,
        upstream_node: Node,
        order: int,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
        start_timing: Optional[TimeRangeBoundary] = None,
        end_timing: Optional[TimeRangeBoundary] = None,
    ) -> Layer:
        """Creates a layer into this time series.

        Args:
            upstream_node: the node whose data will flow into this time series
            order: integer specifying priority of this layer when combining multiple layers; lower order implies
                higher precedence
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to this time
                series
            label_indexer: the slice of the upstream data to pass to this time series
            start_timing: the beginning of the natural time range this layer should produce data for; if None, there
                is no restriction on the beginning of the range
            end_timing: the end of the natural time range this layer should produce data for; if None, there is no
                restriction on the end of the range

        Returns:
            the newly created layer
        """
        return Layer.create(
            downstream_node=self,
            upstream_node=upstream_node,
            order=order,
            output_index=output_index,
            label_indexer=label_indexer,
            start_timing=start_timing,
            end_timing=end_timing,
        )

    def list_layers(self) -> List[Layer]:
        """Lists all layers into this time series."""
        time_series_layers = list_time_series_layers.request_sync(client=get_client(), time_series_uuid=str(self.uuid))

        # Parse the layers into their final representation.
        unsorted_layers = [Layer.parse_obj(layer.dict()) for layer in time_series_layers.data]

        # The API returns the layers ordered by create time, descending. We prefer explicit layer ordering here.
        return sorted(unsorted_layers, key=lambda layer: layer.order)

    def create_run_policy(
        self,
        schedule_timing: Union[Time, TimeDelta, ScheduleTiming],
        start_timing: Optional[TimeRangeBoundary] = None,
        end_timing: Optional[TimeRangeBoundary] = None,
        active: bool = True,
    ) -> TimeSeriesRunPolicy:
        time_series_run_policy_create = TimeSeriesRunPolicyCreate(
            object="Policy",
            type="TimeSeriesRunPolicy",
            schedule_timing=to_timing_create(from_schedule_specifier(schedule_timing)),
            start_timing=to_timing_create(from_time_range_boundary(start_timing)),
            end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            active=active,
        )

        time_series_run_policy = create_time_series_run_policy.request_sync(
            client=get_client(), time_series_uuid=str(self.uuid), json_body=time_series_run_policy_create
        )

        return TimeSeriesRunPolicy.parse_obj(time_series_run_policy.dict())

    def list_run_policies(self) -> List[TimeSeriesRunPolicy]:
        time_series_run_policies = list_time_series_run_policies.request_sync(
            client=get_client(), time_series_uuid=str(self.uuid)
        )

        return [
            TimeSeriesRunPolicy.parse_obj(time_series_run_policy.dict())
            for time_series_run_policy in time_series_run_policies.data
        ]



from typing import List

from myst.models import base_model

from ..models.model_get import ModelGet


class ModelList(base_model.BaseModel):
    """Schema for model list responses."""

    data: List[ModelGet]
    has_more: bool



from myst.client import Client

from ...models.operation_get import OperationGet
from ...models.operation_update import OperationUpdate


def request_sync(client: Client, uuid: str, json_body: OperationUpdate) -> OperationGet:
    """Updates an operation."""

    return client.request(
        method="patch", path=f"/operations/{uuid}", response_class=OperationGet, request_model=json_body
    )



from myst.client import Client

from ...models.model_create import ModelCreate
from ...models.model_get import ModelGet


def request_sync(client: Client, json_body: ModelCreate) -> ModelGet:
    """Creates a model."""

    return client.request(method="post", path=f"/models/", response_class=ModelGet, request_model=json_body)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['myst',
 'myst.adapters',
 'myst.auth',
 'myst.connectors',
 'myst.connectors.model_connectors',
 'myst.connectors.operation_connectors',
 'myst.connectors.source_connectors',
 'myst.core',
 'myst.core.data',
 'myst.core.time',
 'myst.data',
 'myst.models',
 'myst.openapi',
 'myst.openapi.api',
 'myst.openapi.api.model_connectors',
 'myst.openapi.api.models',
 'myst.openapi.api.operation_connectors',
 'myst.openapi.api.operations',
 'myst.openapi.api.organizations',
 'myst.openapi.api.projects',
 'myst.openapi.api.source_connectors',
 'myst.openapi.api.sources',
 'myst.openapi.api.time_series',
 'myst.openapi.api.users',
 'myst.openapi.models',
 'myst.recipes',
 'myst.recipes.model_recipes',
 'myst.recipes.time_series_recipes',
 'myst.resources']

package_data = \
{'': ['*']}

install_requires = \
['google-auth-oauthlib>=0.4.1,<1.0.0',
 'google-auth>=1.11.0,<2.0.0',
 'httpx>=0.15.4,<0.19.0',
 'importlib-metadata',
 'numpy>=1.18.0,<2.0.0',
 'pandas>=0.25.0,<2',
 'pydantic>=1.8.2,<2.0.0',
 'toml>=0.10.2,<0.11.0',
 'typing-extensions>=3.7.2,<4.0.0',
 'urllib3>=1.24.3,<2.0.0']

entry_points = \
{'console_scripts': ['generate-openapi-client = '
                     'tools.generate_openapi_client:app']}

setup_kwargs = {
    'name': 'myst-alpha',
    'version': '0.2.2',
    'description': 'This is the official Python library for the Myst Platform.',
    'long_description': '# Myst Python Library\n\nThis is the official Python client library for the Myst Platform.\n\n## Requirements\n\n- Python 3.7+\n\n## Installation\n\nTo install the package from PyPI:\n\n    $ pip install --upgrade myst-alpha\n\n## Authentication\n\nThe Myst API uses JSON Web Tokens (JWTs) to authenticate requests.\n\nThe Myst Python library handles the sending of JWTs to the API automatically and currently supports two ways to authenticate to obtain a JWT: through your Google user account or a Myst service account.\n\n### Authenticating using your user account\n\nIf you don\'t yet have a Google account, you can create one on the [Google Account Signup](https://accounts.google.com/signup) page.\n\nOnce you have access to a Google account, send an email to `support@myst.ai` with your email so we can authorize you to use the Myst Platform.\n\nUse the following code snippet to authenticate using your user account:\n\n```python\nimport myst\n\nmyst.authenticate()\n```\n\nThe first time you run this, you\'ll be presented with a web browser and asked to authorize the Myst Python library to make requests on behalf of your Google user account.\n\n### Authenticating using a service account\n\nYou can also authenticate using a Myst service account. To request a service account, email `support@myst.ai`.\n\nTo authenticate using a service account, set the `MYST_APPLICATION_CREDENTIALS` environment variable to the path to your service account key file:\n\n```sh\n$ export MYST_APPLICATION_CREDENTIALS=</path/to/key/file.json>\n```\n\nThen, go through the service account authentication flow:\n\n```python\nimport myst\n\nmyst.authenticate_with_service_account()\n```\n\nAlternatively, you can explicitly pass the path to your service account key:\n\n```python\nfrom pathlib import Path\n\nimport myst\n\nmyst.authenticate_with_service_account(key_file_path=Path("/path/to/key/file.json"))\n```\n\n## Working with time series\n\nTime series are at the core of Myst\'s API. To retrieve a time series by UUID:\n\n```python\nimport myst\n\n# You\'ll have to replace this UUID with the UUID of a time series you own.\ntime_series = myst.TimeSeries.get("ca2a63d1-3515-47b4-afc7-13c6656dd744")\n```\n\nYou can insert a `TimeArray` into the time series:\n\n```python\nimport myst\nimport numpy as np\n\ntime_array = myst.TimeArray(\n    sample_period="PT1H",\n    start_time="2021-07-01T00:00:00Z",\n    end_time="2021-07-08T00:00:00Z",\n    as_of_time="2021-07-01T00:00:00Z",\n    values=np.random.randn(168),\n)\ntime_series.insert_time_array(time_array=time_array)\n```\n\nYou can also query a time series for a given as of time and natural time range. In this\nexample, the query should return the data we just inserted:\n\n```python\nimport myst\nfrom myst.testing import assert_time_array_equal\n\nreturned_time_array = time_series.query_time_array(\n    start_time=myst.Time("2021-07-01T00:00:00Z"),\n    end_time=myst.Time("2021-07-08T00:00:00Z"),\n    as_of_time=myst.Time("2021-07-01T00:00:00Z"),\n)\nassert_time_array_equal(returned_time_array, time_array)\n```\n\n## Support\n\nFor questions or just to say hi, reach out to `support@myst.ai`.\n',
    'author': 'Myst AI, Inc.',
    'author_email': 'support@myst.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



from myst.client import Client

from ...models.time_series_create import TimeSeriesCreate
from ...models.time_series_get import TimeSeriesGet


def request_sync(client: Client, json_body: TimeSeriesCreate) -> TimeSeriesGet:
    """Creates a time series."""

    return client.request(method="post", path=f"/time_series/", response_class=TimeSeriesGet, request_model=json_body)



from myst.client import Client

from ...models.source_create import SourceCreate
from ...models.source_get import SourceGet


def request_sync(client: Client, json_body: SourceCreate) -> SourceGet:
    """Creates a source."""

    return client.request(method="post", path=f"/sources/", response_class=SourceGet, request_model=json_body)



from myst.client import Client

from ...models.model_list import ModelList


def request_sync(client: Client) -> ModelList:
    """Lists models."""

    return client.request(method="get", path=f"/models/", response_class=ModelList)



from myst.client import Client

from ...models.model_fit_result_get import ModelFitResultGet


def request_sync(client: Client, model_uuid: str, uuid: str) -> ModelFitResultGet:
    """Gets a model fit result by its unique identifier."""

    return client.request(
        method="get", path=f"/models/{model_uuid}/model_fit_results/{uuid}", response_class=ModelFitResultGet
    )



from myst.client import Client

from ...models.model_run_result_list import ModelRunResultList


def request_sync(client: Client, model_uuid: str) -> ModelRunResultList:
    """Lists model run results for the given model."""

    return client.request(
        method="get", path=f"/models/{model_uuid}/model_run_results/", response_class=ModelRunResultList
    )



from typing import Optional
from uuid import UUID

from myst.openapi.models.deploy_status import DeployStatus
from myst.resources.resource import ShareableResource


class Node(ShareableResource):
    """A node in a project graph.

    Attributes:
        project: identifier of the project to which this node belongs
        title: the title of this node
        description: a brief description of the node
        deploy_status: whether this node is new, deployed, or inactive
    """

    project: UUID
    title: str
    description: Optional[str] = None
    deploy_status: DeployStatus



from myst.client import Client

from ...models.input_list import InputList


def request_sync(client: Client, model_uuid: str) -> InputList:
    """Lists inputs for a model."""

    return client.request(method="get", path=f"/models/{model_uuid}/inputs/", response_class=InputList)



from typing import TYPE_CHECKING, List, Optional, Union

from myst.adapters.timing import to_timing_create
from myst.client import get_client
from myst.connectors.model_connector import ModelConnector
from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta
from myst.models.timing import ScheduleTiming, TimeRangeBoundary, from_schedule_specifier, from_time_range_boundary
from myst.models.types import ItemOrSlice, UUIDOrStr, to_uuid
from myst.openapi.api.models import (
    create_model,
    create_model_fit_policy,
    get_model,
    list_model_fit_policies,
    list_model_inputs,
)
from myst.openapi.models.model_create import ModelCreate
from myst.openapi.models.model_fit_policy_create import ModelFitPolicyCreate
from myst.resources.connector_node import ConnectorNode
from myst.resources.input import Input
from myst.resources.policy import ModelFitPolicy
from myst.resources.time_series import TimeSeries

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.project import Project


class Model(ConnectorNode):
    """A node that learns its parameters during a training phase, and produces output during a prediction phase."""

    @classmethod
    def create(
        cls, project: "Project", title: str, connector: ModelConnector, description: Optional[str] = None
    ) -> "Model":
        """Creates a new model node.

        Args:
            project: the project in which to create the model
            title: the title of the model
            connector: the model connector to use in the model node
            description: a brief description of the model

        Returns:
            the newly created model
        """
        model = create_model.request_sync(
            client=get_client(),
            json_body=ModelCreate(
                object="Node",
                type="Model",
                title=title,
                description=description,
                project=str(project.uuid),
                connector_uuid=str(connector.uuid),
                parameters=connector.parameters_exclude_none(),
            ),
        )

        return Model.parse_obj(model.dict())

    @classmethod
    def get(cls, uuid: UUIDOrStr) -> "Model":
        """Gets a specific model by its identifier."""
        model = get_model.request_sync(client=get_client(), uuid=str(to_uuid(uuid)))

        return Model.parse_obj(model.dict())

    def create_input(
        self,
        upstream_node: TimeSeries,
        group_name: str,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
    ) -> Input:
        """Creates an input into this model.

        Args:
            upstream_node: the time series to feed into this model
            group_name: the name of the input group on this model's connector to which to pass the data from this input
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to this model
            label_indexer: the slice of the upstream data to pass to this model

        Returns:
            the newly created input
        """
        return Input.create(
            downstream_node=self,
            upstream_node=upstream_node,
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

    def list_inputs(self) -> List[Input]:
        """Lists all inputs into this model."""
        model_inputs = list_model_inputs.request_sync(client=get_client(), model_uuid=str(self.uuid))

        return [Input.parse_obj(input_.dict()) for input_ in model_inputs.data]

    def create_fit_policy(
        self,
        schedule_timing: Union[Time, TimeDelta, ScheduleTiming],
        start_timing: Optional[TimeRangeBoundary] = None,
        end_timing: Optional[TimeRangeBoundary] = None,
        active: bool = True,
    ) -> ModelFitPolicy:
        model_fit_policy_create = ModelFitPolicyCreate(
            object="Policy",
            type="ModelFitPolicy",
            schedule_timing=to_timing_create(from_schedule_specifier(schedule_timing)),
            start_timing=to_timing_create(from_time_range_boundary(start_timing)),
            end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            active=active,
        )

        model_fit_policy = create_model_fit_policy.request_sync(
            client=get_client(), model_uuid=str(self.uuid), json_body=model_fit_policy_create
        )

        return ModelFitPolicy.parse_obj(model_fit_policy.dict())

    def list_fit_policies(self) -> List[ModelFitPolicy]:
        model_fit_policies = list_model_fit_policies.request_sync(client=get_client(), model_uuid=str(self.uuid))

        return [ModelFitPolicy.parse_obj(model_fit_policy.dict()) for model_fit_policy in model_fit_policies.data]



from myst.client import Client

from ...models.project_policies_list import ProjectPoliciesList


def request_sync(client: Client, project_uuid: str) -> ProjectPoliciesList:
    """Lists all policies for a project."""

    return client.request(method="get", path=f"/projects/{project_uuid}/policies/", response_class=ProjectPoliciesList)



from myst.client import Client

from ...models.input_create import InputCreate
from ...models.input_get import InputGet


def request_sync(client: Client, model_uuid: str, json_body: InputCreate) -> InputGet:
    """Creates an input for a model."""

    return client.request(
        method="post", path=f"/models/{model_uuid}/inputs/", response_class=InputGet, request_model=json_body
    )



from myst.client import Client

from ...models.input_list import InputList


def request_sync(client: Client, operation_uuid: str) -> InputList:
    """Lists inputs for an operation."""

    return client.request(method="get", path=f"/operations/{operation_uuid}/inputs/", response_class=InputList)



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model

from ..models.deploy_status import DeployStatus


class ProjectGet(base_model.BaseModel):
    """Schema for project get responses."""

    object_: Literal["Project"] = Field(..., alias="object")
    uuid: str
    create_time: str
    organization: str
    owner: str
    title: str
    creator: str
    deploy_status: DeployStatus
    update_time: Optional[str] = None
    description: Optional[str] = None



from typing import List, Union

from myst.models import base_model

from ..models.model_fit_result_get import ModelFitResultGet
from ..models.model_run_result_get import ModelRunResultGet
from ..models.operation_run_result_get import OperationRunResultGet
from ..models.source_run_result_get import SourceRunResultGet
from ..models.time_series_insert_result_get import TimeSeriesInsertResultGet
from ..models.time_series_run_result_get import TimeSeriesRunResultGet


class ProjectResultList(base_model.BaseModel):
    """Project result list schema."""

    data: List[
        Union[
            SourceRunResultGet,
            ModelRunResultGet,
            OperationRunResultGet,
            TimeSeriesRunResultGet,
            TimeSeriesInsertResultGet,
            ModelFitResultGet,
        ]
    ]
    has_more: bool



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class ProjectCreate(base_model.BaseModel):
    """Schema for project create requests."""

    object_: Literal["Project"] = Field(..., alias="object")
    title: str
    description: Optional[str] = None



from myst.client import Client

from ...models.operation_connector_get import OperationConnectorGet


def request_sync(client: Client, uuid: str) -> OperationConnectorGet:
    """Gets an operation connector by its unique identifier."""

    return client.request(method="get", path=f"/operation_connectors/{uuid}", response_class=OperationConnectorGet)



from typing import TYPE_CHECKING, Optional

from myst.client import get_client
from myst.connectors.source_connector import SourceConnector
from myst.models.types import UUIDOrStr, to_uuid
from myst.openapi.api.sources import create_source, get_source
from myst.openapi.models.source_create import SourceCreate
from myst.resources.connector_node import ConnectorNode

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.project import Project


class Source(ConnectorNode):
    """A node which produces data without any inputs."""

    @classmethod
    def create(
        cls, project: "Project", title: str, connector: SourceConnector, description: Optional[str] = None
    ) -> "Source":
        """Creates a new source node.

        Args:
            project: the project in which to create the source
            title: the title of the source
            connector: the source connector to use in the source node
            description: a brief description of the source

        Returns:
            the newly created source
        """
        source = create_source.request_sync(
            client=get_client(),
            json_body=SourceCreate(
                object="Node",
                type="Source",
                title=title,
                description=description,
                project=str(project.uuid),
                connector_uuid=str(connector.uuid),
                parameters=connector.parameters_exclude_none(),
            ),
        )

        return Source.parse_obj(source.dict())

    @classmethod
    def get(cls, uuid: UUIDOrStr) -> "Source":
        """Gets a specific source by its identifier."""
        source = get_source.request_sync(client=get_client(), uuid=str(to_uuid(uuid)))

        return Source.parse_obj(source.dict())



from myst.connectors.operation_connectors.numerical_expression import NumericalExpression
from myst.connectors.operation_connectors.time_transformations import TimeTransformations

__all__ = ["NumericalExpression", "TimeTransformations"]



from typing import List

from myst.models import base_model

from ..models.time_series_run_result_get import TimeSeriesRunResultGet


class TimeSeriesRunResultList(base_model.BaseModel):
    """Schema for time series run result list responses."""

    data: List[TimeSeriesRunResultGet]
    has_more: bool



from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class DeploymentCreate(base_model.BaseModel):
    """Schema for deployment create requests."""

    object_: Literal["Deployment"] = Field(..., alias="object")
    title: str



from abc import ABC, abstractmethod
from typing import Optional

from myst.core.time.time_delta import TimeDelta
from myst.models.types import AxisLabels, CoordinateLabels, Shape
from myst.resources.project import Project
from myst.resources.time_series import TimeSeries


class TimeSeriesRecipe(ABC):
    @abstractmethod
    def create_time_series(
        self,
        project: Project,
        title: str,
        sample_period: TimeDelta,
        cell_shape: Shape = (),
        coordinate_labels: CoordinateLabels = (),
        axis_labels: AxisLabels = (),
        description: Optional[str] = None,
    ) -> TimeSeries:
        raise NotImplementedError()
